/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_GETOPT_HPP
#define OAPR_GETOPT_HPP

#include "oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include "oapr_pool.hpp"
#include "oapr_exception.hpp"
#include "apr_getopt.h"

FILE_VERSION("$Revision: 1.12 $",oapr_getopt_hpp);

namespace oapr
{
  typedef apr_getopt_option_t static_opts;
  //struct static_args: public apr_getopt_option_t {};
  
  #define END_OPTS {NULL, 0, 0, NULL}

  /**
   * Display standard usage information based on the given static args.
   * @param args Static argument definition structure to show usage for.
   * @param cmd Name of command used to invoke this application. Needed for usage. Usually argv[0].
   */
  void usage(const static_opts args[], const char *cmd);

  /**\todo argument support as well as options */
  /**\todo oa framework-style config reader */
  
  class cmdline_processor
  {
  public:
    /**
     * Create a command-line processor.
     * @param argc Number of command-line arguments
     * @param argv Array of pointers to command-line arguments (if own, must be compatible with main() args).
     * @param usepool Pool to use
     * @param arg_definition Static argument definition (struct static_args[])
     * @param allow_interleave Pass true to support interleaving options with regular args
     */
    cmdline_processor(const pool::pool_ptr& p,
		      int argc, const char * const *argv,
		      const static_opts arg_definition[],
		      bool allow_interleave = false,
		      bool throw_on_bad_args = true);

      ~cmdline_processor() {}

      /**
       * Get the first or next option from the command line.
       * @retval Returns false when there are no options left to parse
       * @remark Remember to check the state of the current option, it could be invalid as well as correct.
       */
      bool getopt();

      /**
       * Returns the state of the option selected with a call to getopt(). Check this if getopt() returns true,
       * and you have disabled throwing exceptions on bad args in the constructor.
       *
       * @retval There are three potential status values. They are:
       * <PRE>
       *             APR_BADCH    --  Found a bad option character
       *             APR_BADARG   --  No required argument followed the option
       *             APR_SUCCESS  --  Option found
       * </PRE>
       * @warning Only use this if you disabled throwing exceptions on bad args in the constructor.
       */
      apr_status_t state() { return m_state; }

      /**
       * Option character as returned from getopt().
       */
      char opt() { return m_opt; }
      /**
       * Option argument where applicable, as returned from getopt().
       */
      std::string arg() { std::string out; if(m_arg) out = m_arg; return out; }
      int arg_int() { if(m_arg) return atoi(m_arg); else return 0; }

    private:
      apr_getopt_t* m_os;
      const static_opts *m_opts;
      apr_status_t m_state;
      int m_opt;
      const char *m_arg;
      bool m_throw;
      pool::pool_ptr m_pool;
  };

};

#endif
