/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 Justin Kirby
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 */
#ifndef OAPR_IO_SOCKET_HPP
#define OAPR_IO_SOCKET_HPP
#include "oapr/oapr.hpp"
#include "oapr/oapr_io.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include "apr_network_io.h"
#include "oapr/oapr_io_socketinfo.hpp"
FILE_VERSION("$Revision: 1.4 $",oapr_io_socket_hpp);

namespace oapr
{
  /**
   * basic socket management class
   * Creates and maintians IO to a socket
   */
  class LIBOAPR_API basic_socket : public basic_io
  {
  public:
    typedef pool_mgr::pool_t pool_t;
    typedef pool_mgr::pool_ptr pool_ptr;
  public:
    basic_socket(const pool_ptr &p)
      : basic_io(p),
	m_sock(0),
	m_remote(0),
	m_local(0),
	m_socktype(0),
	m_created(false),
	m_isopen(false)
    {}
    /**
     * Constructs a socket based from another one. Usually used
     * by accept.
     * @param sock The apr socket to assign to this
     * @param p The apr pool managing the socket
     * @param type The type of the socket.
     */
    basic_socket(apr_socket_t *sock, const basic_io::pool_ptr& p,int type)
      : basic_io(p),
	m_sock(sock),
	m_remote(0),
	m_local(0),
	m_socktype(type),/**todo this is an error, need to get it*/
	m_created(true),
	m_isopen(true)
    {}
    basic_socket(const basic_socket& s)
      : basic_io(s),
	m_sock(s.m_sock),
	m_remote(s.m_remote),
	m_local(s.m_local),
	m_socktype(s.m_socktype),
	m_created(s.m_created),
	m_isopen(s.m_isopen)
    {}
    virtual ~basic_socket(){}

    /**
     * Creates a socket
     * @param family The family 
     * @param type The type
     * @return result of the create
     */
    virtual apr_status_t create(int family, int type);

    /**
     * Shutdown a socket cleanly
     * @param how How to shutdown the socket,
     * - APR_SHUTDOWN_READ no longer allow read requests
     * - APR_SHUTDOWN_WRITE no longer allow write requests
     * - APR_SHUTDOWN_READWRITE no longer allow read or writes requests
     * @return result of the shutdown
     */
    virtual apr_status_t shutdown(apr_shutdown_how_e how);

    /**
     * Bind the socket to a set ip & port
     * @param sa created sockinfo
     * @return result of bind 
     * @throw oapr::logic_error if the socket is not created
     */
    virtual apr_status_t bind(const sockinfo &sa);

    /**
     * Listen to a bound socket for connections
     * @param backlog Number of outstanding connections to allow
     * for the listen queue.
     * @throw oapr::logic_error is the socket is not open
     */
    virtual apr_status_t listen(apr_int32_t backlog);

    /**
     * Accept a new connection request
     * @return A new constructed socket
     * @throw oapr::logic_error if the socket is not created and opened
     * @throw oapr::oapr_exception if the pool for the new socket 
     * fails to create
     * @throw oapr::oapr_exception if the call to apr_socket_accept fails
     * @throw oapr::exception_base will be rethrown if the ctor of the
     * new socket throws
     */
    virtual basic_socket accept();

    /**
     * Issue a connection request to a socket.
     * @param si created sockinfo
     * @throw oapr::logic_error if the socket is not created
     */
    virtual apr_status_t connect(const sockinfo &si);

    /**
     * Setup socket options 
     * @param opt Option to configure, one of:
     * - APR_SO_DEBUF     -- turn on debugging information
     * - APR_SO_KEEPALIVE -- keep connections active
     * - APR_SO_LINGER    -- lingers on close if data is present
     * - APR_SO_NONBLOCK  -- Turns blocking on/off 
     * - APR_SO_REUSEADDR -- The rules used in validating addresses
     *                       supplied to bind should allow reuse
     * - APR_SO_SNDBUF    -- Set the Send buffer size
     * - APR_SO_RCVBUF    -- Set the receive buffer size
     * @param on Value for the option
     * @throw oapr::logic_error if the socket is not created
     */
    virtual apr_status_t option(apr_int32_t opt, apr_int32_t on);

    /**
     * Check whether the socket is created
     * @return true is the socket is created, false otherwise
     */
    virtual bool is_created()const{return m_created;}

    /**
     * Close the socket entirely
     * This performs an orderly shutdown of the socket and then
     * closes it
     * @return result of the close
     */
    virtual apr_status_t close();

    /**
     * This is a no-op that will always return false
     */
    virtual bool eof();

    /**
     * Check the status of the socket
     * If the socket has been bound or connected this will return true
     */
    virtual bool is_open()const{return m_isopen;}

    /**
     * Write data out to socket
     * @param buff buffer to write
     * @param toWrite number of bytes to send
     * @param full wether to force sending of entire buffer.
     * @return number of bytes written
     * @throw oapr::logic_error if the socket is not open
     * @throw oapr::invalid_arg if the type is not of 
     * - SOCK_STREAM
     * - SOCK_DGRAM
     * @throw oapr::oapr_exception if the call to send fails 
     */
    virtual apr_size_t write(const void* buff,
			     apr_size_t toWrite,
			     bool full=false);

    /**
     * Read data from socket
     * @param buff buffer to read data into
     * @param toRead number of bytes to read
     * @param full whether to force fill the buffer
     * @return number of bytes read
     * @throw oapr::logic_error if the socket is not created and open
     * @throw oapr::invalid_arg if the the socket is not of
     * - SOCK_STREAM
     * - SOCK_DGRAM
     * @throw oapr::oapr_exception if the call to read fails
     */
    virtual apr_size_t read(void* buff,
			    apr_size_t toRead,
			    bool full=false)const;


    /**
     * currently a no-op
     */
    virtual apr_status_t flush();

    /**
     * Setup socket timeout 
     * @param Value for the timeout
     * @return result of the set
     * @throw oapr::logic_error if the socket is not created
     */
    virtual apr_status_t timeout(apr_interval_time_t timeout);
    /**
     * GEt the timeout of the socket
     * @return the timeout of the socket
     * @throw oapr::logic_error if the socket is not created
     */
    virtual apr_interval_time_t timeout()const;

  protected:
  private:
    apr_socket_t *m_sock;/**< The actual socket*/
    apr_sockaddr_t *m_remote;/**< remote socket addr*/
    apr_sockaddr_t *m_local;/**< local socket addr*/
    int m_socktype;/**< the type of the socket*/
    bool m_created;/**< status flag whether the socket has been created*/
    bool m_isopen;/**< whether the socket is connected/bound*/
  };
};
#endif
