/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_PROCESS_HPP
#define OAPR_PROCESS_HPP

#include "oapr/oapr.hpp"
#include "oapr/oapr_buildconfig.hpp"
#include "oapr_exception.hpp"
#include "oapr_pool.hpp"
#include "apr_thread_proc.h"
#include <string>
#include <vector>
#include <assert.h>
#include "boost/shared_ptr.hpp"
FILE_VERSION("$Revision: 1.2 $",oapr_process_hpp);

namespace oapr {

  class process
  {
  };

  class LIBOAPR_API child_process: public process
  {
    public:
      /**
       * Create process object.
       * @param ap Shared pool to use.
       */
      child_process(pool& ap);
      /**
       * Create and start process object.
       * @param progname The program to run
       * @param ap Shared pool to use.
       * @param args the arguments to pass to the new program.  The first
       *             one should be the program name.
       * @param env The new environment table for the new process.  This
       *            should be a list of NULL-terminated strings. This argument
       *            is ignored for APR_PROGRAM_ENV and APR_PROGRAM_PATH types
       *            of commands.
       */
      child_process(pool& ap, const std::string& progname, std::vector<std::string>& args, const std::vector<std::string>& env);
      ~child_process();

      // Local functions
      //static void fork(process **process);

      inline void kill(int sig) { apr_proc_kill(&m_proc, sig); }

      /**
       * Wait for a child process to die
       * @param exitcode The returned exit status of the child, if a child process 
       *                 dies, or the signal that caused the child to die.
       *                 On platforms that don't support obtaining this information, 
       *                 throws a not_implemented error.
       * Parameter is optional and may be NULL.
       * @param exitwhy Why the child died, the bitwise or of:
       * <PRE>
       *            APR_PROC_EXIT         -- process terminated normally
       *            APR_PROC_SIGNAL       -- process was killed by a signal
       *            APR_PROC_SIGNAL_CORE  -- process was killed by a signal, and
       *                                     generated a core dump.
       * </PRE>
       * Parameter is optional and may be NULL.
       */
      void waitfor(int *exitcode = NULL, apr_exit_why_e *exitwhy = NULL);
      /**
       * check if a child process has died.
       * @param exitcode The returned exit status of the child, if a child process 
       *                 dies, or the signal that caused the child to die.
       *                 On platforms that don't support obtaining this information, 
       *                 throws a not_implemented error.
       * Parameter is optional and may be NULL.
       * @param exitwhy Why the child died, the bitwise or of:
       * <PRE>
       *            APR_PROC_EXIT         -- process terminated normally
       *            APR_PROC_SIGNAL       -- process was killed by a signal
       *            APR_PROC_SIGNAL_CORE  -- process was killed by a signal, and
       *                                     generated a core dump.
       * </PRE>
       * Parameter is optional and may be NULL.
       */
      bool done(int *exitcode = NULL, apr_exit_why_e *exitwhy = NULL);
    protected:
      void create_proc(const std::string& progname, std::vector<std::string>& args, const std::vector<std::string>& env);
    private:
      boost::shared_ptr<pool> m_pool;
      std::string m_progname;
      apr_proc_t m_proc;
      bool m_assigned;
  };

  class LIBOAPR_API this_process: public process
  {
    enum daemon_option_t { foreground=0, background=1 };
    public:
      /**
       * Detach the process from the controlling terminal.
       * @param daemonize set to 'background' (default) if the process should daemonize
       *                  and become a background process, or 'foreground' if the
       *                  process should stay in the foreground.
       */
      apr_status_t detach(daemon_option_t daemonize = background);
      /**
       * Execute a standard unix fork.
       */
       /**\todo non-portable, should not support.*/
      apr_status_t fork(process &new_process);
  };
  
#if APR_HAS_FORK
/**
 * This is currently the only non-portable call in APR.  This executes 
 * a standard unix fork.
 * @param proc The resulting process handle. 
 * @param cont The pool to use. 
 */
//APR_DECLARE(apr_status_t) fork_process(apr_proc_t *proc, apr_pool_t *cont);
#endif

};

#endif
