/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 Matt Carter. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

#ifndef OAPR_SQLTOOLS_HPP
#define OAPR_SQLTOOLS_HPP

#include <string>
#include <sstream>
#include <vector>

namespace oapr
{
  /**
   * These SQL Builder classes have limited functionality so far - i.e. what I've needed.
   * Where clauses are always coupled with "AND" and you must take care of your own
   * quoting/escaping of values as per your platform, e.g. pqxx::Quote()
   */
  
  class sql_builder
  {
    public:
      sql_builder() {}
      virtual ~sql_builder() {}
      virtual void set(const std::string& field, const std::string& value) = 0;
      virtual void set(const std::string& field, int value) = 0;
      virtual void set(const std::string& field, double value) = 0;
      virtual std::string str() = 0;
      virtual void clear() = 0;
      void table(const char *table) { m_table = table; }
      void table(const std::string& table) { m_table = table; }
    protected:
      std::string m_table;
  };
  
  class sql_select_builder;
  
  class sql_insert_builder: public sql_builder
  {
    public:
      sql_insert_builder(): m_setcount(0) {}
      ~sql_insert_builder() {}
      void set(const std::string& field, const std::string& value)
      {
        if(value.empty() || field.empty()) return;
        if(m_setcount != 0)
        {
          m_fields << ",";
          m_values << ",";
        }
        m_fields << field;
        m_values << value;
        ++m_setcount;
      }
      void set(const std::string& field, int value)
      {
        if(m_setcount != 0)
        {
          m_fields << ",";
          m_values << ",";
        }
        m_fields << field;
        m_values << value;
        ++m_setcount;
      }
      void set(const std::string& field, double value)
      {
        if(m_setcount != 0)
        {
          m_fields << ",";
          m_values << ",";
        }
        m_fields << field;
        m_values << value;
        ++m_setcount;
      }
      std::string str() {
        std::ostringstream sql;
        if(m_setcount == 0) return sql.str();
        sql << "INSERT INTO " << m_table << " (" << m_fields.str() << ") VALUES (" << m_values.str() << ")";
        return sql.str();
      }
      void clear() {
        m_setcount = 0;
        m_fields.str("");
        m_values.str("");
        m_table.clear();
      }
    private:
      std::ostringstream m_fields;
      std::ostringstream m_values;
      int m_setcount;
  };

  class sql_update_builder: public sql_builder
  {
    public:
      sql_update_builder(): m_setcount(0), m_wherecount(0) {}
      ~sql_update_builder() {}
      void set(const std::string& field, const std::string& value)
      {
        if(value.empty() || field.empty()) return;
        if(m_setcount == 0)
        {
          m_set << "SET " << field << " = " << value;
        }
        else
        {
          m_set << ", " << field << " = " << value;
        }
        ++m_setcount;
      }
      void set(const std::string& field, int value)
      {
        if(m_setcount == 0)
        {
          m_set << "SET " << field << " = " << value;
        }
        else
        {
          m_set << ", " << field << " = " << value;
        }
        ++m_setcount;
      }
      void set(const std::string& field, double value)
      {
        if(m_setcount == 0)
        {
          m_set << "SET " << field << " = " << value;
        }
        else
        {
          m_set << ", " << field << " = " << value;
        }
        ++m_setcount;
      }
      void where(const std::string& condition)
      {
        if(m_wherecount == 0)
        {
          m_where << condition;
        }
        else
        {
          m_where << " AND " << condition;
        }
        ++m_wherecount;
      }
      template<typename T> inline void where(const std::string& field, const char *oper, T &value)
      {
        if(m_wherecount == 0)
        {
          m_where << field << " " << oper << " " << value;
        }
        else
        {
          m_where << " AND " << field << " " << oper << " " << value;
        }
        ++m_wherecount;
      }
      template<typename T> inline void where(sql_select_builder& select, const char *oper, T& value)
      {
        if(m_wherecount == 0)
        {
          m_where << "(" << select << ") " << oper << " " << value;
        }
        else
        {
          m_where << " AND (" << select << ") " << oper << " " << value;
        }
        ++m_wherecount;
      }
      void clear() {
        m_setcount = 0;
        m_wherecount = 0;
        m_set.str("");
        m_where.str("");
        m_table.clear();
      }
      std::string str() {
        std::ostringstream sql;
        if(m_setcount == 0 || m_wherecount == 0) return sql.str();
        sql << "UPDATE " << m_table << " " << m_set.str() << " WHERE " << m_where.str();
        return sql.str();
      }
      
    private:
      std::ostringstream m_set;
      std::ostringstream m_where;
      int m_setcount;
      int m_wherecount;
  };

  class sql_delete_builder: public sql_builder
  {
    public:
      sql_delete_builder(): m_wherecount(0) {}
      ~sql_delete_builder() {}
      
      /**\todo stubs .. required by base class. fix */
      void set(const std::string& field, const std::string& value) {}
      void set(const std::string& field, int value) {}
      void set(const std::string& field, double value) {}
      
      void where(const std::string& condition)
      {
        if(m_wherecount == 0)
        {
          m_where << condition;
        }
        else
        {
          m_where << " AND " << condition;
        }
        ++m_wherecount;
      }
      template<typename T> inline void where(const std::string& field, const char *oper, T& value)
      {
        if(m_wherecount == 0)
        {
          m_where << field << " " << oper << " " << value;
        }
        else
        {
          m_where << " AND " << field << " " << oper << " " << value;
        }
        ++m_wherecount;
      }
      template<typename T> inline void where(sql_select_builder& select, const char *oper, T& value)
      {
        if(m_wherecount == 0)
        {
          m_where << "(" << select << ") " << oper << " " << value;
        }
        else
        {
          m_where << " AND (" << select << ") " << oper << " " << value;
        }
        ++m_wherecount;
      }
      void clear() {
        m_wherecount = 0;
        m_where.str("");
        m_table.clear();
      }
      std::string str() {
        std::ostringstream sql;
        if(m_wherecount == 0) return sql.str();
        sql << "DELETE FROM " << m_table << " WHERE " << m_where.str();
        return sql.str();
      }
      
    private:
      std::ostringstream m_where;
      int m_wherecount;
  };

  class sql_select_builder: public sql_builder
  {
    public:
      sql_select_builder(): m_fieldcount(0), m_wherecount(0) {}
      ~sql_select_builder() {}
      
      /**\todo stubs .. required by base class. fix */
      void set(const std::string& field, const std::string& value) {}
      void set(const std::string& field, int value) {}
      void set(const std::string& field, double value) {}
      
      void get(const std::string& field)
      {
        if(field.empty()) return;
        if(m_fieldcount == 0)
        {
          m_fields << field;
        }
        else
        {
          m_fields << ", " << field;
        }
        ++m_fieldcount;
      }
      void where(const std::string& condition)
      {
        if(m_wherecount == 0)
        {
          m_where << condition;
        }
        else
        {
          m_where << " AND " << condition;
        }
        ++m_wherecount;
      }
      template<typename T> inline void where(const std::string& field, const char *oper, T &value)
      {
        if(m_wherecount == 0)
        {
          m_where << field << " " << oper << " " << value;
        }
        else
        {
          m_where << " AND " << field << " " << oper << " " << value;
        }
        ++m_wherecount;
      }
      void clear() {
        m_fieldcount = 0;
        m_wherecount = 0;
        m_fields.str("");
        m_where.str("");
        m_table.clear();
      }
      std::string str() {
        std::ostringstream sql;
        sql << "SELECT ";
        if(m_fieldcount == 0) sql << "*"; else sql << m_fields.str();
        sql << " FROM " << m_table;
        if(m_wherecount != 0) sql << " WHERE " << m_where.str();
        return sql.str();
      }
    private:
      std::ostringstream m_fields;
      std::ostringstream m_where;
      int m_fieldcount;
      int m_wherecount;
  };
  
  class sql_multiplexer
  {
    public:
      typedef std::vector<sql_builder*> builder_ptr_list;
    public:
      sql_multiplexer() {}
      ~sql_multiplexer() {}
      void add_builder(sql_builder& builder)
      {
        m_builders.push_back(&builder);
      }
      void set(const std::string& field, const std::string& value)
      {
        for(builder_ptr_list::const_iterator i = m_builders.begin(); i != m_builders.end(); ++i) (*i)->set(field, value);
      }
      void set(const std::string& field, int value)
      {
        for(builder_ptr_list::const_iterator i = m_builders.begin(); i != m_builders.end(); ++i) (*i)->set(field, value);
      }
      void set(const std::string& field, double value)
      {
        for(builder_ptr_list::const_iterator i = m_builders.begin(); i != m_builders.end(); ++i) (*i)->set(field, value);
      }
      void clear()
      {
        for(builder_ptr_list::const_iterator i = m_builders.begin(); i != m_builders.end(); ++i) (*i)->clear();
      }
      void table(const char *tablename)
      {
        for(builder_ptr_list::const_iterator i = m_builders.begin(); i != m_builders.end(); ++i) (*i)->table(tablename);
      }
      void table(const std::string& tablename)
      {
        for(builder_ptr_list::const_iterator i = m_builders.begin(); i != m_builders.end(); ++i) (*i)->table(tablename);
      }
    private:
      builder_ptr_list m_builders;
  };
  
};

#endif
